searchData={"items":[{"type":"module","title":"TypedStructor","doc":"`TypedStructor` is a library for defining structs with types effortlessly.\n(This library is a rewritten version of [TypedStruct](https://github.com/ejpcmac/typed_struct) because it is no longer actively maintained.)","ref":"TypedStructor.html"},{"type":"module","title":"Installation - TypedStructor","doc":"Add `:typed_structor` to the list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:typed_structor, \"~> 0.4\"}\n  ]\nend\n```\n\nAdd `:typed_structor` to your `.formatter.exs` file\n\n```elixir\n[\n  # import the formatter rules from `:typed_structor`\n  import_deps: [..., :typed_structor],\n  inputs: [...]\n]\n```","ref":"TypedStructor.html#module-installation"},{"type":"module","title":"Usage - TypedStructor","doc":"","ref":"TypedStructor.html#module-usage"},{"type":"module","title":"General usage - TypedStructor","doc":"To define a struct with types, use `TypedStructor`,\nand then define fields under the `TypedStructor.typed_structor/2` macro,\nusing the `TypedStructor.field/3` macro to define each field.\n\n```elixir\ndefmodule User do\n  # use TypedStructor to import the `typed_structor` macro\n  use TypedStructor\n\n  typed_structor do\n    # Define each field with the `field` macro.\n    field :id, pos_integer()\n\n    # set a default value\n    field :name, String.t(), default: \"Unknown\"\n\n    # enforce a field\n    field :age, non_neg_integer(), enforce: true\n  end\nend\n```\nThis is equivalent to:\n```elixir\ndefmodule User do\n  defstruct [:id, :name, :age]\n\n  @type t() :: %__MODULE__{\n    id: pos_integer() | nil,\n    # Note: The 'name' can not be nil, for it has a default value.\n    name: String.t(),\n    age: non_neg_integer()\n  }\nend\n```\nCheck `TypedStructor.typed_structor/2` and `TypedStructor.field/3` for more information.","ref":"TypedStructor.html#module-general-usage"},{"type":"module","title":"Options - TypedStructor","doc":"You can also generate an `opaque` type for the struct,\neven changing the type name:\n\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor type_kind: :opaque, type_name: :profile do\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer()\n  end\nend\n```\nThis is equivalent to:\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  defstruct [:id, :name, :age]\n\n  @opaque profile() :: %__MODULE__{\n    id: pos_integer() | nil,\n    name: String.t() | nil,\n    age: non_neg_integer() | nil\n  }\nend\n```\n\nType parameters also can be defined:\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor do\n    parameter :id\n    parameter :name\n\n    field :id, id\n    field :name, name\n    field :age, non_neg_integer()\n  end\nend\n```\nbecomes:\n```elixir\ndefmodule User do\n  @type t(id, name) :: %__MODULE__{\n    id: id | nil,\n    name: name | nil,\n    age: non_neg_integer() | nil\n  }\n\n  defstruct [:id, :name, :age]\nend\n```\n\nIf you prefer to define a struct in a submodule, you can use\nthe `module` option with `TypedStructor`. This allows you to\nencapsulate the struct definition within a specific submodule context.\n\nConsider this example:\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  # `%User.Profile{}` is generated\n  typed_structor module: Profile do\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer()\n  end\nend\n```\nWhen defining a struct in a submodule, the `typed_structor` block\nfunctions similarly to a `defmodule` block. Therefore,\nthe previous example can be alternatively written as:\n```elixir\ndefmodule User do\n  defmodule Profile do\n    use TypedStructor\n\n    typed_structor do\n      field :id, pos_integer()\n      field :name, String.t()\n      field :age, non_neg_integer()\n    end\n  end\nend\n```\n\nFurthermore, the `typed_structor` block allows you to\ndefine functions, derive protocols, and more, just\nas you would within a `defmodule` block. Here's a example:\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor module: Profile, define_struct: false do\n    @derive {Jason.Encoder, only: [:email]}\n    field :email, String.t()\n\n    use Ecto.Schema\n    @primary_key false\n\n    schema \"users\" do\n      Ecto.Schema.field(:email, :string)\n    end\n\n    import Ecto.Changeset\n\n    def changeset(%__MODULE__{} = user, attrs) do\n      user\n      |> cast(attrs, [:email])\n      |> validate_required([:email])\n    end\n  end\nend\n```\nNow, you can interact with these structures:\n```elixir\niex> User.Profile.__struct__()\n%User.Profile{__meta__: #Ecto.Schema.Metadata<:built, \"users\">, email: nil}\niex> Jason.encode!(%User.Profile{})\n\"{\\\"email\\\":null}\"\niex> User.Profile.changeset(%User.Profile{}, %{\"email\" => \"my@email.com\"})\n#Ecto.Changeset ,\n  valid?: true\n>\n```","ref":"TypedStructor.html#module-options"},{"type":"module","title":"Define an Exception - TypedStructor","doc":"In Elixir, an exception is defined as a struct that includes a special field named `__exception__`.\nTo define an exception, use the `defexception` definer within the `typed_structor` block.\n\n```elixir\ndefmodule HTTPException do\n  use TypedStructor\n\n  typed_structor definer: :defexception, enforce: true do\n    field :status, non_neg_integer()\n  end\n\n  @impl Exception\n  def message(%__MODULE__{status: status}) do\n    \"HTTP status #{status}\"\n  end\nend\n```","ref":"TypedStructor.html#module-define-an-exception"},{"type":"module","title":"Define records related macros - TypedStructor","doc":"In Elixir, you can use the Record module to define and work with Erlang records,\nmaking interoperability between Elixir and Erlang more seamless.\n\n```elixir\ndefmodule TypedStructor.User do\n  use TypedStructor\n\n  typed_structor definer: :defrecord, record_name: :user, record_tag: User, enforce: true do\n    field :name, String.t()\n    field :age, pos_integer()\n  end\nend\n```","ref":"TypedStructor.html#module-define-records-related-macros"},{"type":"module","title":"Documentation - TypedStructor","doc":"To add a `@typedoc` to the struct type, just add the attribute in the typed_structor block:\n\n```elixir\ntyped_structor do\n  @typedoc \"A typed user\"\n\n  field :id, pos_integer()\n  field :name, String.t()\n  field :age, non_neg_integer()\nend\n```\nYou can also document submodules this way:\n\n```elixir\ntypedstructor module: Profile do\n  @moduledoc \"A user profile struct\"\n  @typedoc \"A typed user profile\"\n\n  field :id, pos_integer()\n  field :name, String.t()\n  field :age, non_neg_integer()\nend\n```","ref":"TypedStructor.html#module-documentation"},{"type":"module","title":"Plugins - TypedStructor","doc":"`TypedStructor` offers a plugin system to enhance functionality.\nFor details on creating a plugin, refer to the `TypedStructor.Plugin` module.\n\nHere is a example of `Guides.Plugins.Accessible` plugin to define `Access` behavior for the struct.\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor do\n    plugin Guides.Plugins.Accessible\n\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer()\n  end\nend\n\nuser = %User{id: 1, name: \"Phil\", age: 20}\nget_in(user, [:name]) # => \"Phil\"\n```\n\n> #### Plugins guides {: .tip}\n>\n> Here are some [Plugin Guides](guides/plugins/introduction.md)\n> for creating your own plugins. Please check them out\n> and feel free to copy-paste the code.","ref":"TypedStructor.html#module-plugins"},{"type":"macro","title":"TypedStructor.field/3","doc":"Defines a field in a `typed_structor/2`.\nYou can override the options set by `typed_structor/2` by passing options.","ref":"TypedStructor.html#field/3"},{"type":"macro","title":"Example - TypedStructor.field/3","doc":"# A field named :example of type String.t()\n    field :example, String.t()","ref":"TypedStructor.html#field/3-example"},{"type":"macro","title":"Options - TypedStructor.field/3","doc":"* `:default` - sets the default value for the field\n  * `:enforce` - if set to `true`, enforces the field, and makes its type\n    non-nullable if `:default` is not set\n\n>","ref":"TypedStructor.html#field/3-options"},{"type":"macro","title":"How `:default` and `:enforce` affect `type` and `@enforce_keys` {: .tip} - TypedStructor.field/3","doc":">\n> | **`:default`** | **`:enforce`** | **`type`**        | **`@enforce_keys`** |\n> | -------------- | -------------- | ----------------- | ------------------- |\n> | `set`          | `true`         | `t()`             | `excluded`          |\n> | `set`          | `false`        | `t()`             | `excluded`          |\n> | `unset`        | `true`         | `t()`             | **`included`**      |\n> | `unset`        | `false`        | **`t() \\| nil`** | `excluded`          |","ref":"TypedStructor.html#field/3-how-default-and-enforce-affect-type-and-enforce_keys-tip"},{"type":"macro","title":"TypedStructor.parameter/2","doc":"Defines a type parameter in a `typed_structor/2`.","ref":"TypedStructor.html#parameter/2"},{"type":"macro","title":"Example - TypedStructor.parameter/2","doc":"# A type parameter named int\n    parameter :int\n\n    fied :number, int # not int()","ref":"TypedStructor.html#parameter/2-example"},{"type":"macro","title":"TypedStructor.plugin/2","doc":"Registers a plugin for the currently defined struct.","ref":"TypedStructor.html#plugin/2"},{"type":"macro","title":"Example - TypedStructor.plugin/2","doc":"typed_structor do\n      plugin MyPlugin\n\n      field :string, String.t()\n    end\n\nFor more information on how to define your own plugins, please see\n`TypedStructor.Plugin`. To use a third-party plugin, please refer directly to\nits documentation.","ref":"TypedStructor.html#plugin/2-example"},{"type":"macro","title":"TypedStructor.typed_structor/2","doc":"Defines a struct with type information.\n\nInside a `typed_structor` block, you can define fields with the `field/3` macro.","ref":"TypedStructor.html#typed_structor/2"},{"type":"macro","title":"Options - TypedStructor.typed_structor/2","doc":"* `:module` - if provided, a new submodule will be created with the struct.\n  * `:enforce` - if `true`, the struct will enforce the keys, see `field/3` options for more information.\n  * `:definer` - the definer module to use to define the struct, record or exception. Defaults to `:defstruct`. It also accepts a macro that receives the definition struct and returns the AST. See definer section below.\n  * `:type_kind` - the kind of type to use for the struct. Defaults to `type`, can be `opaque` or `typep`.\n  * `:type_name` - the name of the type to use for the struct. Defaults to `t`.","ref":"TypedStructor.html#typed_structor/2-options"},{"type":"macro","title":"Definer - TypedStructor.typed_structor/2","doc":"The available definers are:\n- `:defstruct`, which defines a struct and a type for a given definition\n- `:defexception`, which defines an exception and a type for a given definition\n- `:defrecord`, which defines record macros and a type for a given definition\n- `:defrecordp`, which defines private record macros and a type for a given definition\n\n### `:defstruct` options\n\n* `:define_struct` - if `false`, the type will be defined, but the struct will not be defined. Defaults to `true`.\n\n\n### `:defexception` options\n\n* `:define_struct` - if `false`, the type will be defined, but the struct will not be defined. Defaults to `true`.\n\n\n### `:defrecord` and `:defrecordp` options\n\n* `:record_name`(**required**) - the name of the record, it must be provided.\n* `:record_tag` - if set, the record will be tagged with the given value. Defaults to `nil`.\n* `:define_record` - if `false`, the type will be defined, but the record will not be defined. Defaults to `true`.","ref":"TypedStructor.html#typed_structor/2-definer"},{"type":"macro","title":"custom definer - TypedStructor.typed_structor/2","doc":"defmodule MyStruct do\n      use TypedStructor\n\n      typed_structor definer: MyDefiner do\n\n        field :name, String.t()\n        field :age, integer()\n      end\n    end","ref":"TypedStructor.html#typed_structor/2-custom-definer"},{"type":"macro","title":"Examples - TypedStructor.typed_structor/2","doc":"defmodule MyStruct do\n      use TypedStructor\n\n      typed_structor do\n        field :name, String.t()\n        field :age, integer()\n      end\n    end\n\nCreates the struct in a submodule instead:\n\n    defmodule MyStruct do\n      use TypedStructor\n\n      typed_structor module: Struct do\n        field :name, String.t()\n        field :age, integer()\n      end\n    end\n\nTo add a `@typedoc` to the struct type and `@moduledoc` to the submodule,\njust add the module attribute in the `typed_structor` block:\n\n    defmodule MyStruct do\n      use TypedStructor\n\n      typed_structor module: Struct do\n        @typedoc \"A typed struct\"\n        @moduledoc \"A submodule\"\n\n        field :name, String.t()\n        field :age, integer()\n      end\n    end","ref":"TypedStructor.html#typed_structor/2-examples"},{"type":"module","title":"TypedStructor.Definition","doc":"The definition struct that holds the `TypedStructor` options,\nfields and parameters.","ref":"TypedStructor.Definition.html"},{"type":"type","title":"TypedStructor.Definition.t/0","doc":"","ref":"TypedStructor.Definition.html#t:t/0"},{"type":"behaviour","title":"TypedStructor.Plugin","doc":"This module defines the plugin behaviour for `TypedStructor`.","ref":"TypedStructor.Plugin.html"},{"type":"behaviour","title":"Plugin Behaviour - TypedStructor.Plugin","doc":"A plugin is a module that implements the `TypedStructor.Plugin` behaviour.\nThree macro callbacks are available for injecting code at different stages:\n\n* `c:init/1`: This macro callback is called when the plugin is used.\n* `c:before_definition/2`: This macro callback is called right before defining the struct.\nNote hat plugins will run in the order they are registered.\n* `c:after_definition/2`: This macro callback is called right after defining the struct.\nNote that plugins will run in the **reverse** order they are registered.\n\n> #### Plugins guides {: .tip}\n>\n> Here are some [Plugin Guides](guides/plugins/introduction.md)\n> for creating your own plugins. Please check them out\n> and feel free to copy-paste the code.","ref":"TypedStructor.Plugin.html#module-plugin-behaviour"},{"type":"behaviour","title":"Example - TypedStructor.Plugin","doc":"Let's define a plugin that defines `Ecto.Schema` while defining a typed struct.\nThis plugin takes a `:source` option which passing to `Ecto.Schema.schema/2`,\nyou can use `belongs_to` and `has_many` directly in the module.\nIt would be used like this:\n```elixir\ndefmodule MyApp do\n  use TypedStructor\n\n  # fix aliases\n  alias __MODULE__.User\n  alias __MODULE__.Post\n\n  typed_structor module: User do\n    # import the plugin with source option\n    plugin EctoSchemaPlugin, source: \"users\"\n\n    field :name, :string, enforce: true\n    field :age, :integer, default: 0\n    # pass redact option to Ecto.Schema.field/3\n    field :password, :string, redact: true\n    has_many :posts, Post\n  end\n\n  typed_structor module: Post do\n    # import the plugin with source option\n    plugin EctoSchemaPlugin, source: \"posts\"\n\n    field :title, :string, enforce: true\n    field :content, :string, enforce: true\n    belongs_to :user, User, enforce: true\n  end\nend\n```\nAfter compiled, you got:\n```elixir\niex> t MyApp.User\n@type t() :: %MyApp.User{\n        __meta__: Ecto.Schema.Metadata.t(),\n        age: integer() | nil,\n        id: integer(),\n        name: String.t(),\n        password: String.t() | nil,\n        posts: [MyApp.Post.t()] | nil\n      }\n\niex> t MyApp.Post\n@type t() :: %MyApp.Post{\n        __meta__: Ecto.Schema.Metadata.t(),\n        content: String.t(),\n        id: integer(),\n        title: String.t(),\n        user: MyApp.User.t(),\n        user_id: integer()\n      }\n\niex> MyApp.User.__schema__(:redact_fields)\n[:password]\n\niex> MyApp.User.__schema__(:association, :posts)\n%Ecto.Association.Has{\n  cardinality: :many,\n  field: :posts,\n  owner: MyApp.User,\n  related: MyApp.Post,\n  owner_key: :id,\n  related_key: :user_id,\n  on_cast: nil,\n  queryable: MyApp.Post,\n  on_delete: :nothing,\n  on_replace: :raise,\n  where: [],\n  unique: true,\n  defaults: [],\n  relationship: :child,\n  ordered: false,\n  preload_order: []\n}\n\niex> MyApp.Post.__schema__(:association, :user)\n%Ecto.Association.BelongsTo{\n  field: :user,\n  owner: MyApp.Post,\n  related: MyApp.User,\n  owner_key: :user_id,\n  related_key: :id,\n  queryable: MyApp.User,\n  on_cast: nil,\n  on_replace: :raise,\n  where: [],\n  defaults: [],\n  cardinality: :one,\n  relationship: :parent,\n  unique: true,\n  ordered: false\n}\n```\nFollowing is the implementation of the plugin:\n```elixir\ndefmodule EctoSchemaPlugin do\n  use TypedStructor.Plugin\n\n  @impl TypedStructor.Plugin\n  defmacro init(opts) do\n    quote do\n      unless Keyword.has_key?(unquote(opts), :source) do\n        raise \"The `:source` option is not provided.\"\n      end\n\n      # import association functions to the module,\n      # so that we can use `has_many` and `belongs_to` directly\n      import unquote(__MODULE__), only: [has_many: 2, belongs_to: 3]\n    end\n  end\n\n  @impl TypedStructor.Plugin\n  defmacro before_definition(definition, _opts) do\n    # manipulate the definition before defining the struct\n    quote do\n      unquote(definition)\n      # disable defining struct, for Ecto.Schema will define it\n      |> Map.update!(:options, &Keyword.put(&1, :define_struct, false))\n      |> Map.update!(:fields, fn fields ->\n        Enum.flat_map(fields, fn field ->\n          {ecto_type, options} = Keyword.pop!(field, :type)\n          type = unquote(__MODULE__).__ecto_type_to_type__(ecto_type)\n\n          field = Keyword.merge(options, type: type, ecto_type: ecto_type)\n\n          case ecto_type do\n            {:belongs_to, name} ->\n              foreign_key_name =\n                name\n                |> Macro.expand(__ENV__)\n                |> Module.split()\n                |> List.last()\n                |> Macro.underscore()\n                |> Kernel.<>(\"_id\")\n                |> String.to_atom()\n\n              foreign_key =\n                Keyword.merge(options, name: foreign_key_name, type: quote(do: integer()))\n\n              [foreign_key, field]\n\n            _other ->\n              [field]\n          end\n        end)\n      end)\n      |> Map.update!(\n        :fields,\n        &[\n          [name: :__meta__, type: quote(do: Ecto.Schema.Metadata.t()), enforce: true],\n          [name: :id, type: quote(do: integer()), enforce: true]\n          | &1\n        ]\n      )\n    end\n  end\n\n  @impl TypedStructor.Plugin\n  defmacro after_definition(definition, opts) do\n    # here we define the Ecto.Schema\n    quote bind_quoted: [definition: definition, opts: opts] do\n      use Ecto.Schema\n\n      source = Keyword.fetch!(opts, :source)\n\n      schema source do\n        for options <- definition.fields do\n          {name, options} = Keyword.pop!(options, :name)\n          {ecto_type, options} = Keyword.pop(options, :ecto_type)\n          options = Keyword.take(options, [:primary_key, :default, :redact])\n\n          case ecto_type do\n            nil ->\n              # skip some fields\n              nil\n\n            {:has_many, module} ->\n              module = Macro.expand(module, __ENV__)\n\n              has_many name, module, options\n\n            {:belongs_to, module} ->\n              module = Macro.expand(module, __ENV__)\n\n              belongs_to name, module, options\n\n            _ ->\n              field name, ecto_type, options\n          end\n        end\n      end\n    end\n  end\n\n  defmacro has_many(name, queryable) do\n    quote do\n      field unquote(name), {:has_many, unquote(queryable)}\n    end\n  end\n\n  defmacro belongs_to(name, queryable, opts) do\n    quote do\n      field unquote(name), {:belongs_to, unquote(queryable)}, unquote(opts)\n    end\n  end\n\n  def __ecto_type_to_type__(:string), do: quote(do: String.t())\n  def __ecto_type_to_type__(:integer), do: quote(do: integer())\n  def __ecto_type_to_type__({:has_many, module}), do: quote(do: [unquote(module).t()])\n  def __ecto_type_to_type__({:belongs_to, module}), do: quote(do: unquote(module).t())\nend\n```","ref":"TypedStructor.Plugin.html#module-example"},{"type":"macrocallback","title":"TypedStructor.Plugin.after_definition/2","doc":"This macro callback is called right after defining the struct.\n\nIt receives the definition of the struct and the plugin options,\nand its return value is ignored.","ref":"TypedStructor.Plugin.html#c:after_definition/2"},{"type":"macrocallback","title":"TypedStructor.Plugin.before_definition/2","doc":"This macro callback is called right before defining the struct.\n\nIt receives the definition of the struct and the plugin options,\nand it should return the `TypedStructor.Definition` struct or\na list which contains exactly one `TypedStructor.Definition` struct.","ref":"TypedStructor.Plugin.html#c:before_definition/2"},{"type":"macrocallback","title":"TypedStructor.Plugin.init/1","doc":"This macro callback is called when the plugin is used.\n\nHere you can define module attributes, import modules, etc.","ref":"TypedStructor.Plugin.html#c:init/1"},{"type":"module","title":"TypedStructor.Definer.Defexception","doc":"A definer to define an exception and a type for a given definition.","ref":"TypedStructor.Definer.Defexception.html"},{"type":"module","title":"Additional options for `typed_structor` - TypedStructor.Definer.Defexception","doc":"* `:define_struct` - if `false`, the type will be defined, but the struct will not be defined. Defaults to `true`.","ref":"TypedStructor.Definer.Defexception.html#module-additional-options-for-typed_structor"},{"type":"module","title":"Usage - TypedStructor.Definer.Defexception","doc":"defmodule MyException do\n      use TypedStructor\n\n      typed_structor definer: :defexception, define_struct: false do\n        field :message, String.t()\n      end\n    end","ref":"TypedStructor.Definer.Defexception.html#module-usage"},{"type":"macro","title":"TypedStructor.Definer.Defexception.define/1","doc":"Defines an exception and a type for a given definition.","ref":"TypedStructor.Definer.Defexception.html#define/1"},{"type":"module","title":"TypedStructor.Definer.Defrecord","doc":"A definer to define record macors and a type for a given definition.","ref":"TypedStructor.Definer.Defrecord.html"},{"type":"module","title":"Additional options for `typed_structor` - TypedStructor.Definer.Defrecord","doc":"* `:record_name`(**required**) - the name of the record, it must be provided.\n* `:record_tag` - if set, the record will be tagged with the given value. Defaults to `nil`.\n* `:define_record` - if `false`, the type will be defined, but the record will not be defined. Defaults to `true`.","ref":"TypedStructor.Definer.Defrecord.html#module-additional-options-for-typed_structor"},{"type":"module","title":"Usage - TypedStructor.Definer.Defrecord","doc":"defmodule MyRecord do\n      use TypedStructor\n\n      typed_structor definer: :defrecord, record_name: :user, record_tag: User, define_recrod: true do\n        field :name, String.t(), enforce: true\n        field :age, pos_integer(), enforce: true\n      end\n    end\n\nThe above code is equivalent to:\n\n    defmodule MyRecord do\n      require Record\n\n      @type t() :: {User, name :: String.t(), age :: pos_integer()}\n\n      Record.defrecord(:user, User, [:name, :age])\n    end","ref":"TypedStructor.Definer.Defrecord.html#module-usage"},{"type":"macro","title":"TypedStructor.Definer.Defrecord.define/1","doc":"Defines an exception and a type for a given definition.","ref":"TypedStructor.Definer.Defrecord.html#define/1"},{"type":"module","title":"TypedStructor.Definer.Defrecordp","doc":"A definer to define private record macros and a type for a given definition.\n\nSee more at `TypedStructor.Definer.Defrecord`.","ref":"TypedStructor.Definer.Defrecordp.html"},{"type":"macro","title":"TypedStructor.Definer.Defrecordp.define/1","doc":"Defines an exception and a type for a given definition.","ref":"TypedStructor.Definer.Defrecordp.html#define/1"},{"type":"module","title":"TypedStructor.Definer.Defstruct","doc":"A definer to define a struct and a type for a given definition.","ref":"TypedStructor.Definer.Defstruct.html"},{"type":"module","title":"Additional options for `typed_structor` - TypedStructor.Definer.Defstruct","doc":"* `:define_struct` - if `false`, the type will be defined, but the struct will not be defined. Defaults to `true`.","ref":"TypedStructor.Definer.Defstruct.html#module-additional-options-for-typed_structor"},{"type":"module","title":"Usage - TypedStructor.Definer.Defstruct","doc":"defmodule MyStruct do\n      use TypedStructor\n\n      typed_structor definer: :defstruct, define_struct: false do\n        field :name, String.t()\n        field :age, integer()\n      end\n    end","ref":"TypedStructor.Definer.Defstruct.html#module-usage"},{"type":"macro","title":"TypedStructor.Definer.Defstruct.define/1","doc":"Defines a struct and a type for a given definition.","ref":"TypedStructor.Definer.Defstruct.html#define/1"},{"type":"module","title":"TypedStructor.Definer.Utils","doc":"Utilities for definer modules.","ref":"TypedStructor.Definer.Utils.html"},{"type":"function","title":"TypedStructor.Definer.Utils.fields_and_enforce_keys/1","doc":"Extracts fields and enforce keys from a definition.","ref":"TypedStructor.Definer.Utils.html#fields_and_enforce_keys/1"},{"type":"function","title":"TypedStructor.Definer.Utils.types/2","doc":"Extracts types from a definition.","ref":"TypedStructor.Definer.Utils.html#types/2"},{"type":"extras","title":"Changelog","doc":"# Changelog\n\n## [0.5.0](https://github.com/elixir-typed-structor/typed_structor/compare/v0.4.2...v0.5.0) (2024-09-27)","ref":"changelog.html"},{"type":"extras","title":"Features - Changelog","doc":"* add `:defexception` denfiner to define an exception ([#19](https://github.com/elixir-typed-structor/typed_structor/issues/19)) ([800ca9e](https://github.com/elixir-typed-structor/typed_structor/commit/800ca9e1e0d145288ccab7469153b65fedb63c0a))\n* support `:defrecord` and `:defrecordp` and custom definer ([#21](https://github.com/elixir-typed-structor/typed_structor/issues/21)) ([cc93185](https://github.com/elixir-typed-structor/typed_structor/commit/cc9318589bd8444a1597b6283478dcafcc61f0e0))\n\n## [0.4.2](https://github.com/elixir-typed-structor/typed_structor/compare/v0.4.1...v0.4.2) (2024-07-11)","ref":"changelog.html#features"},{"type":"extras","title":"Documentation - Changelog","doc":"* this release is a documentation update and does not contain any new features or bug fixes\n\n\n## [0.4.1](https://github.com/elixir-typed-structor/typed_structor/compare/v0.4.0...v0.4.1) (2024-07-10)","ref":"changelog.html#documentation"},{"type":"extras","title":"Documentation - Changelog","doc":"* this release is a documentation update and does not contain any new features or bug fixes\n\n\n## [0.4.0](https://github.com/elixir-typed-structor/typed_structor/compare/v0.3.0...v0.4.0) (2024-07-07)","ref":"changelog.html#documentation"},{"type":"extras","title":"⚠ BREAKING CHANGES - Changelog","doc":"* store parameters as keywords instead of atoms to retain more information ([#10](https://github.com/elixir-typed-structor/typed_structor/issues/10))","ref":"changelog.html#breaking-changes"},{"type":"extras","title":"Code Refactoring - Changelog","doc":"* store parameters as keywords instead of atoms to retain more information ([#10](https://github.com/elixir-typed-structor/typed_structor/issues/10)) ([df5b17e](https://github.com/elixir-typed-structor/typed_structor/commit/df5b17e09cb569cb1c93c470db0e9eb322e429b8))","ref":"changelog.html#code-refactoring"},{"type":"extras","title":"Features - Changelog","doc":"* add doc_fields guide ([#11](https://github.com/elixir-typed-structor/typed_structor/pull/11))\n\n## [0.3.0](https://github.com/elixir-typed-structor/typed_structor/compare/v0.2.0...v0.3.0) (2024-07-03)","ref":"changelog.html#features"},{"type":"extras","title":"⚠ BREAKING CHANGES - Changelog","doc":"* remove accessible plugin, add it to guides","ref":"changelog.html#breaking-changes"},{"type":"extras","title":"Bug Fixes - Changelog","doc":"* macro capture will capture import context ([9c6f340](https://github.com/elixir-typed-structor/typed_structor/commit/9c6f3406009565b1e7ccab77d7ce24a7ee3984e6))","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"Code Refactoring - Changelog","doc":"* remove accessible plugin, add it to guides ([204cb0a](https://github.com/elixir-typed-structor/typed_structor/commit/204cb0aeab5a7ca8d5a15d7e016243d66af6e595))\n\n## [0.2.0](https://github.com/elixir-typed-structor/typed_structor/compare/v0.1.5...v0.2.0) (2024-06-30)","ref":"changelog.html#code-refactoring"},{"type":"extras","title":"Features - Changelog","doc":"* register plugins globally by configuration ([4767e62](https://github.com/elixir-typed-structor/typed_structor/commit/4767e620c237777f535cfe763c773d13a4944c0f))","ref":"changelog.html#features"},{"type":"extras","title":"0.1.5 (2024-06-29) - Changelog","doc":"This release is a documentation update and\ndoes not contain any new features or bug fixes.","ref":"changelog.html#0-1-5-2024-06-29"},{"type":"extras","title":"0.1.4 (2024-06-29) - Changelog","doc":"This release is a documentation update and\ndoes not contain any new features or bug fixes.","ref":"changelog.html#0-1-4-2024-06-29"},{"type":"extras","title":"0.1.3 (2024-06-29) - Changelog","doc":"","ref":"changelog.html#0-1-3-2024-06-29"},{"type":"extras","title":"Features - Changelog","doc":"- doc: add `Migrate from typed_struct` guide","ref":"changelog.html#features"},{"type":"extras","title":"Bug Fixes - Changelog","doc":"- fix: exclude key from `@enforce_keys`, if `:default` is set and `:enforce` is true","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"0.1.2 (2024-06-29) - Changelog","doc":"","ref":"changelog.html#0-1-2-2024-06-29"},{"type":"extras","title":"Bug Fixes - Changelog","doc":"- avoid enforcing the field when default is unset and enforce is false","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"0.1.1 (2024-06-28) - Changelog","doc":"","ref":"changelog.html#0-1-1-2024-06-28"},{"type":"extras","title":"Bug Fixes - Changelog","doc":"- if default is set, then this field is enforced (#2)","ref":"changelog.html#bug-fixes"},{"type":"extras","title":"0.1.0 (2024-06-28) - Changelog","doc":"","ref":"changelog.html#0-1-0-2024-06-28"},{"type":"extras","title":"Features - Changelog","doc":"First release","ref":"changelog.html#features"},{"type":"extras","title":"Introduction","doc":"# TypedStructor\n\n[![Build Status](https://github.com/elixir-typed-structor/typed_structor/actions/workflows/elixir.yml/badge.svg)](https://github.com/elixir-typed-structor/typed_structor/actions/workflows/elixir.yml)\n[![Hex.pm](https://img.shields.io/hexpm/v/typed_structor)](https://hex.pm/packages/typed_structor)\n[![Document](https://img.shields.io/badge/document-gray)](https://hexdocs.pm/typed_structor)\n[![Plugin guides](https://img.shields.io/badge/plugin_guides-indianred?label=%F0%9F%94%A5&labelColor=snow)](https://hexdocs.pm/typed_structor/introduction.html)\n\n<!-- MODULEDOC -->\n\n`TypedStructor` is a library for defining structs with types effortlessly.\n(This library is a rewritten version of [TypedStruct](https://github.com/ejpcmac/typed_struct) because it is no longer actively maintained.)","ref":"readme.html"},{"type":"extras","title":"Installation - Introduction","doc":"Add `:typed_structor` to the list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:typed_structor, \"~> 0.4\"}\n  ]\nend\n```\n\nAdd `:typed_structor` to your `.formatter.exs` file\n\n```elixir\n[\n  # import the formatter rules from `:typed_structor`\n  import_deps: [..., :typed_structor],\n  inputs: [...]\n]\n```","ref":"readme.html#installation"},{"type":"extras","title":"Usage - Introduction","doc":"","ref":"readme.html#usage"},{"type":"extras","title":"General usage - Introduction","doc":"To define a struct with types, use `TypedStructor`,\nand then define fields under the `TypedStructor.typed_structor/2` macro,\nusing the `TypedStructor.field/3` macro to define each field.\n\n```elixir\ndefmodule User do\n  # use TypedStructor to import the `typed_structor` macro\n  use TypedStructor\n\n  typed_structor do\n    # Define each field with the `field` macro.\n    field :id, pos_integer()\n\n    # set a default value\n    field :name, String.t(), default: \"Unknown\"\n\n    # enforce a field\n    field :age, non_neg_integer(), enforce: true\n  end\nend\n```\nThis is equivalent to:\n```elixir\ndefmodule User do\n  defstruct [:id, :name, :age]\n\n  @type t() :: %__MODULE__{\n    id: pos_integer() | nil,\n    # Note: The 'name' can not be nil, for it has a default value.\n    name: String.t(),\n    age: non_neg_integer()\n  }\nend\n```\nCheck `TypedStructor.typed_structor/2` and `TypedStructor.field/3` for more information.","ref":"readme.html#general-usage"},{"type":"extras","title":"Options - Introduction","doc":"You can also generate an `opaque` type for the struct,\neven changing the type name:\n\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor type_kind: :opaque, type_name: :profile do\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer()\n  end\nend\n```\nThis is equivalent to:\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  defstruct [:id, :name, :age]\n\n  @opaque profile() :: %__MODULE__{\n    id: pos_integer() | nil,\n    name: String.t() | nil,\n    age: non_neg_integer() | nil\n  }\nend\n```\n\nType parameters also can be defined:\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor do\n    parameter :id\n    parameter :name\n\n    field :id, id\n    field :name, name\n    field :age, non_neg_integer()\n  end\nend\n```\nbecomes:\n```elixir\ndefmodule User do\n  @type t(id, name) :: %__MODULE__{\n    id: id | nil,\n    name: name | nil,\n    age: non_neg_integer() | nil\n  }\n\n  defstruct [:id, :name, :age]\nend\n```\n\nIf you prefer to define a struct in a submodule, you can use\nthe `module` option with `TypedStructor`. This allows you to\nencapsulate the struct definition within a specific submodule context.\n\nConsider this example:\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  # `%User.Profile{}` is generated\n  typed_structor module: Profile do\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer()\n  end\nend\n```\nWhen defining a struct in a submodule, the `typed_structor` block\nfunctions similarly to a `defmodule` block. Therefore,\nthe previous example can be alternatively written as:\n```elixir\ndefmodule User do\n  defmodule Profile do\n    use TypedStructor\n\n    typed_structor do\n      field :id, pos_integer()\n      field :name, String.t()\n      field :age, non_neg_integer()\n    end\n  end\nend\n```\n\nFurthermore, the `typed_structor` block allows you to\ndefine functions, derive protocols, and more, just\nas you would within a `defmodule` block. Here's a example:\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor module: Profile, define_struct: false do\n    @derive {Jason.Encoder, only: [:email]}\n    field :email, String.t()\n\n    use Ecto.Schema\n    @primary_key false\n\n    schema \"users\" do\n      Ecto.Schema.field(:email, :string)\n    end\n\n    import Ecto.Changeset\n\n    def changeset(%__MODULE__{} = user, attrs) do\n      user\n      |> cast(attrs, [:email])\n      |> validate_required([:email])\n    end\n  end\nend\n```\nNow, you can interact with these structures:\n```elixir\niex> User.Profile.__struct__()\n%User.Profile{__meta__: #Ecto.Schema.Metadata<:built, \"users\">, email: nil}\niex> Jason.encode!(%User.Profile{})\n\"{\\\"email\\\":null}\"\niex> User.Profile.changeset(%User.Profile{}, %{\"email\" => \"my@email.com\"})\n#Ecto.Changeset ,\n  valid?: true\n>\n```","ref":"readme.html#options"},{"type":"extras","title":"Define an Exception - Introduction","doc":"In Elixir, an exception is defined as a struct that includes a special field named `__exception__`.\nTo define an exception, use the `defexception` definer within the `typed_structor` block.\n\n```elixir\ndefmodule HTTPException do\n  use TypedStructor\n\n  typed_structor definer: :defexception, enforce: true do\n    field :status, non_neg_integer()\n  end\n\n  @impl Exception\n  def message(%__MODULE__{status: status}) do\n    \"HTTP status #{status}\"\n  end\nend\n```","ref":"readme.html#define-an-exception"},{"type":"extras","title":"Define records related macros - Introduction","doc":"In Elixir, you can use the Record module to define and work with Erlang records,\nmaking interoperability between Elixir and Erlang more seamless.\n\n```elixir\ndefmodule TypedStructor.User do\n  use TypedStructor\n\n  typed_structor definer: :defrecord, record_name: :user, record_tag: User, enforce: true do\n    field :name, String.t()\n    field :age, pos_integer()\n  end\nend\n```","ref":"readme.html#define-records-related-macros"},{"type":"extras","title":"Documentation - Introduction","doc":"To add a `@typedoc` to the struct type, just add the attribute in the typed_structor block:\n\n```elixir\ntyped_structor do\n  @typedoc \"A typed user\"\n\n  field :id, pos_integer()\n  field :name, String.t()\n  field :age, non_neg_integer()\nend\n```\nYou can also document submodules this way:\n\n```elixir\ntypedstructor module: Profile do\n  @moduledoc \"A user profile struct\"\n  @typedoc \"A typed user profile\"\n\n  field :id, pos_integer()\n  field :name, String.t()\n  field :age, non_neg_integer()\nend\n```","ref":"readme.html#documentation"},{"type":"extras","title":"Plugins - Introduction","doc":"`TypedStructor` offers a plugin system to enhance functionality.\nFor details on creating a plugin, refer to the `TypedStructor.Plugin` module.\n\nHere is a example of `Guides.Plugins.Accessible` plugin to define `Access` behavior for the struct.\n```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor do\n    plugin Guides.Plugins.Accessible\n\n    field :id, pos_integer()\n    field :name, String.t()\n    field :age, non_neg_integer()\n  end\nend\n\nuser = %User{id: 1, name: \"Phil\", age: 20}\nget_in(user, [:name]) # => \"Phil\"\n```\n\n> #### Plugins guides {: .tip}\n>\n> Here are some [Plugin Guides](guides/plugins/introduction.md)\n> for creating your own plugins. Please check them out\n> and feel free to copy-paste the code.","ref":"readme.html#plugins"},{"type":"extras","title":"Migrate from typed_struct","doc":"# Migrate from `typed_struct`\n\nThis package(`typed_structor`) keeps the same API as `typed_struct`.\nIt is a drop-in replacement for `typed_struct`.","ref":"migrate_from_typed_struct.html"},{"type":"extras","title":"Migration - Migrate from typed_struct","doc":"1. Replace `typed_struct` with `typed_structor` in your `mix.exs` file.\n\n```diff\n   defp deps do\n     [\n       # ...deps\n-      {:typed_struct, \"~> 0.3.0\"},\n+      {:typed_structor, \"~> 0.4\"},\n     ]\n   end\n```\n\n2. Run `mix do deps.unlock --unused, deps.get, deps.clean --unused` to fetch the new dependency.\n3. Replace `TypedStruct` with `TypedStructor` in your code.\n\n```diff\n defmodule User do\n-  use TypedStruct\n+  use TypedStructor\n \n-  typedstruct do\n+  typed_structor do\n     field :id, pos_integer()\n     field :name, String.t()\n     field :age, non_neg_integer()\n   end\n end\n```\n4. That's it!","ref":"migrate_from_typed_struct.html#migration"},{"type":"extras","title":"Introduction","doc":"# Introduction to the plugin system\n\nFor more customization, `TypedStructor` provides a plugin system\nthat allows you to extend the functionality of the library.\nThis is useful when you want to extract some common logic into a separate module.\n\nSee `TypedStructor.Plugin` for how to create a plugin.\n\nThis library comes with a few built-in plugins, and we don't like to\nimplement more built-in plugins, but instead, we encourage you to create your own plugins.\nWe provide some example plugins that you can use as a reference, or copy-paste.\n\n**Plugin examples:**\n- [Registering plugins globally](./registering_plugins_globally.md)\n- [Implement `Access` behavior](./accessible.md)\n- [Implement reflection functions](./reflection.md)\n- [Type Only on Ecto Schema](./type_only_on_ecto_schema.md)\n- [Add primary key and timestamps types to your Ecto schema](./primary_key_and_timestamps.md)\n- [Derives the `Jason.Encoder` for `struct`](./derive_jason.md)\n- [Derives the `Enumerable` for `struct`](./derive_enumerable.md)","ref":"introduction.html"},{"type":"extras","title":"Registering plugins globally","doc":"# Registering plugins globally\n\n`TypedStructor` allows you to register plugins globally, so you don't have to specify them in each struct.\nThis is useful when you want to apply the same plugin to all modules that use `TypedStructor`.\n\n> #### Global plugins are applied to all modules {: .warning}\n> The global registered plugins is applied to **all modules** that use `TypedStructor`.\n> That means any dependency or library that uses `TypedStructor` will also be affected.\n>\n> If you want to apply the plugin to specific modules, you can determine the\n> module name or other conditions in the plugin implementation.","ref":"registering_plugins_globally.html"},{"type":"extras","title":"Usage - Registering plugins globally","doc":"To register a plugin globally, you can add it to the `:plugins` key in the `:typed_structor` app configuration.\n```elixir\nconfig :typed_structor, plugins: [MyPlugin, {MyPluginWithOpts, [foo: :bar]}]\n```","ref":"registering_plugins_globally.html#usage"},{"type":"extras","title":"How to opt-out the plugin conditionally - Registering plugins globally","doc":"The most common way to opt-out a plugin is to determine the module name.\n\n```elixir\ndefmodule MyPlugin do\n  use TypedStructor.Plugin\n\n  @impl TypedStructor.Plugin\n  defmacro before_definition(definition, _opts) do\n    quote do\n      if unquote(__MODULE__).__opt_in__?(__MODULE__) do\n        # do something\n      end\n\n      unquote(definition)\n    end\n  end\n\n  @impl TypedStructor.Plugin\n  defmacro after_definition(_definition, _opts) do\n    quote do\n      if unquote(__MODULE__).__opt_in__?(__MODULE__) do\n        # do something\n      end\n    end\n  end\n\n  def __opt_in__?(module) when is_atom(module) do\n    # Opt-in only for schemas under MyApp.Schemas\n    String.starts_with?(Atom.to_string(module), \"MyApp.Schemas\")\n  end\nend\n```","ref":"registering_plugins_globally.html#how-to-opt-out-the-plugin-conditionally"},{"type":"extras","title":"Implement Access behavior","doc":"# Implement `Access` behavior\n\nSometimes, you may want to use the `get_in/2` and `update_in/3` functions\nwith your structs. This guide demonstrates how to effortlessly implement\nthe `Access` behavior for your structs using a plugin.","ref":"accessible.html"},{"type":"extras","title":"Implementation - Implement Access behavior","doc":"> #### Destructive operations  {: .warning}\n> These operations are not allowed for the struct:\n> * update `:__struct__` key\n> * pop a key\n>\n> The functions will raise an `ArgumentError` if called.\n> To enable these functionalities, override the `get_and_update/3` and `pop/2` functions.\n\n```elixir\ndefmodule Guides.Plugins.Accessible do\n  @moduledoc \"\"\"\n  This plugin implements the `Access` behavior for the struct\n  by delegating the `fetch/2`, `get_and_update/3`, and `pop/2`\n  functions to the `Map` module.\n\n  > #### Destructive operations  {: .warning}\n  > These operations are not allowed for the struct:\n  > * update `:__struct__` key\n  > * pop a key\n  >\n  > The functions will raise an `ArgumentError` if called.\n  > To enable these functionalities, override the `get_and_update/3` and `pop/2` functions.","ref":"accessible.html#implementation"},{"type":"extras","title":"Usage - Implement Access behavior","doc":"typed_structor do\n        plugin TypedStructor.Plugins.Accessible\n\n        # fields\n      end\n  \"\"\"\n\n  use TypedStructor.Plugin\n\n  @impl TypedStructor.Plugin\n  defmacro after_definition(_definition, _opts) do\n    quote do\n      @behaviour Access\n\n      @impl Access\n      defdelegate fetch(term, key), to: Map\n\n      @impl Access\n      def get_and_update(data, :__struct__, _function) do\n        raise ArgumentError,\n              \"Cannot update `:__struct__` key.\" <>\n                \"To enable this functionality, implement `Access.get_and_update/3` for #{inspect(__MODULE__)} to override this behaviour.\"\n      end\n\n      defdelegate get_and_update(data, key, function), to: Map\n\n      @impl Access\n      def pop(data, key) do\n        raise ArgumentError,\n              \"Cannot pop `#{inspect(key)}` key.\\n\" <>\n                \"To enable this functionality, implement `Access.pop/2` for #{inspect(__MODULE__)} to override this behaviour.\"\n      end\n\n      @defoverridable Access\n    end\n  end\nend\n```","ref":"accessible.html#usage"},{"type":"extras","title":"Usage - Implement Access behavior","doc":"```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor do\n    plugin Guides.Plugins.Accessible\n\n    field :name, String.t()\n    field :age, integer()\n  end\nend\n```\n\n```elixir\niex> user = %User{name: \"Phil\", age: 20}\n%User{name: \"Phil\", age: 20}\niex> get_in(user, [:name])\n\"Phil\"\niex> put_in(user, [:name], \"phil\")\n%User{name: \"phil\", age: 20}\n```","ref":"accessible.html#usage"},{"type":"extras","title":"Implement reflection functions","doc":"# Implement reflection functions\n\nDefine a plugin that generates reflection functions that\ncan be used to access the fields and parameters of a struct.","ref":"reflection.html"},{"type":"extras","title":"Implement - Implement reflection functions","doc":"```elixir\ndefmodule Guides.Plugins.Reflection do\n  use TypedStructor.Plugin\n\n  @impl TypedStructor.Plugin\n  defmacro after_definition(definition, _opts) do\n    quote bind_quoted: [definition: definition] do\n      fields = Enum.map(definition.fields, &Keyword.fetch!(&1, :name))\n\n      enforced_fields =\n        definition.fields\n        |> Stream.filter(fn field ->\n          Keyword.get_lazy(field, :enforce, fn ->\n            Keyword.get(definition.options, :enforce, false)\n          end)\n        end)\n        |> Stream.map(&Keyword.fetch!(&1, :name))\n        |> Enum.to_list()\n\n      def __typed_structor__(:fields), do: unquote(fields)\n      def __typed_structor__(:parameters), do: Enum.map(unquote(definition.parameters), &Keyword.fetch!(&1, :name))\n      def __typed_structor__(:enforced_fields), do: unquote(enforced_fields)\n\n      for field <- definition.fields do\n        name = Keyword.fetch!(field, :name)\n        type = field |> Keyword.fetch!(:type) |> Macro.escape()\n\n        def __typed_structor__(:type, unquote(name)), do: unquote(type)\n        def __typed_structor__(:field, unquote(name)), do: unquote(Macro.escape(field))\n      end\n    end\n  end\nend\n```","ref":"reflection.html#implement"},{"type":"extras","title":"Usage - Implement reflection functions","doc":"```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor do\n    plugin Guides.Plugins.Reflection\n\n    parameter :age\n\n    field :name, String.t(), enforce: true\n    field :age, age, default: 20\n  end\nend\n\ndefmodule MyApp do\n  use TypedStructor\n\n  typed_structor module: User, enforce: true do\n    plugin Guides.Plugins.Reflection\n\n    field :name, String.t()\n    field :age, integer()\n  end\nend\n```\n\n```elixir\niex> User.__typed_structor__(:fields)\n[:name, :age]\niex> User.__typed_structor__(:parameters)\n[:age]\n\niex> MyApp.User.__typed_structor__(:enforced_fields)\n[:name, :age]\niex> Macro.to_string(User.__typed_structor__(:type, :age))\n\"age\"\n```","ref":"reflection.html#usage"},{"type":"extras","title":"Add fields docs to the @typedoc","doc":"# Add fields docs to the `@typedoc`\n\nFor enhanced readability, it’s often beneficial to document the fields\nand parameters of a struct. Incorporating documentation can significantly\naid both your co-workers and your future self by providing clear,\naccessible information.\n\nBelow is the final code generated by the plugin, which includes comprehensive\ndocumentation for the `User` struct:\n\n```\ndefmodule User do\n  @typedoc \"\"\"\n  @type t(age) :: %User{age: age | nil, name: String.t() | nil}\n\n  This is a user struct.","ref":"doc_fields.html"},{"type":"extras","title":"Parameters - Add fields docs to the @typedoc","doc":"Name | Description\n  :age | The age parameter.","ref":"doc_fields.html#parameters"},{"type":"extras","title":"Fields - Add fields docs to the @typedoc","doc":"Name  | Type             | Description\n  :name | String.t() | nil | The name of the user.\n  :age  | age | nil        | The age of the user.\n  \"\"\"\n  @type t(age) :: %__MODULE__{name: String.t() | nil, age: age | nil}\n\n  defstruct [:name, :age]\nend\n```","ref":"doc_fields.html#fields"},{"type":"extras","title":"Implement - Add fields docs to the @typedoc","doc":"```elixir\ndefmodule Guides.Plugins.DocFields do\n  @moduledoc \"\"\"\n  The `DocFields` plugin generates documentation for fields and parameters.\n  Simply add the `:doc` option to the `field` and `parameter` macros to document them.","ref":"doc_fields.html#implement"},{"type":"extras","title":"Example - Add fields docs to the @typedoc","doc":"use TypedStructor\n\n      typed_structor do\n        @typedoc \\\"\"\"\n        This is a user struct.\n        \\\"\"\"\n        plugin Guides.Plugins.DocFields\n\n        parameter :age, doc: \"The age parameter.\"\n\n        field :name, String.t(), doc: \"The name of the user.\"\n        field :age, age, doc: \"The age of the user.\"\n      end\n\n  This will generate the following documentation for you:\n\n      @typedoc \\\"\"\"\n      This is a user struct.","ref":"doc_fields.html#example"},{"type":"extras","title":"Parameters - Add fields docs to the @typedoc","doc":"| Name | Description |\n      |------|-------------|\n      |`:age` | The age parameter.|","ref":"doc_fields.html#parameters"},{"type":"extras","title":"Fields - Add fields docs to the @typedoc","doc":"| Name | Type | Description |\n      |------|------|-------------|\n      |`:name` | `String.t() \\| nil` | The name of the user.|\n      |`:age` | `age \\| nil` | The age of the user.|\n      \\\"\"\"\n\n      @type t(age) :: %User{age: age | nil, name: String.t() | nil}\n  \"\"\"\n\n  use TypedStructor.Plugin\n\n  @impl TypedStructor.Plugin\n  defmacro before_definition(definition, _opts) do\n    quote do\n      @typedoc unquote(__MODULE__).__generate_doc__(unquote(definition), @typedoc)\n\n      unquote(definition)\n    end\n  end\n\n  def __generate_doc__(_definition, false), do: nil\n\n  def __generate_doc__(definition, typedoc) do\n    parameters =\n      Enum.map(definition.parameters, fn parameter ->\n        name = Keyword.fetch!(parameter, :name)\n        doc = Keyword.get(parameter, :doc, \"*not documented*\")\n\n        [\"`#{inspect(name)}`\", doc]\n      end)\n\n    parameters_docs =\n      if length(parameters) > 0 do\n        \"\"\"","ref":"doc_fields.html#fields"},{"type":"extras","title":"Parameters - Add fields docs to the @typedoc","doc":"| Name | Description |\n        |------|-------------|\n        #{join_rows(parameters)}\n        \"\"\"\n      end\n\n    fields =\n      Enum.map(definition.fields, fn field ->\n        name = Keyword.fetch!(field, :name)\n\n        type = Keyword.fetch!(field, :type)\n\n        type =\n          if Keyword.get(field, :enforce, false) or Keyword.has_key?(field, :default) do\n            Macro.to_string(type)\n          else\n            # escape `|`\n            \"#{Macro.to_string(type)} \\\\| nil\"\n          end\n\n        doc = Keyword.get(field, :doc, \"*not documented*\")\n\n        [\"`#{inspect(name)}`\", \"`#{type}`\", doc]\n      end)\n\n    fields_docs =\n      if length(fields) > 0 do\n        \"\"\"","ref":"doc_fields.html#parameters"},{"type":"extras","title":"Fields - Add fields docs to the @typedoc","doc":"| Name | Type | Description |\n        |------|------|-------------|\n        #{join_rows(fields)}\n        \"\"\"\n      end\n\n    [parameters_docs, fields_docs]\n    |> Enum.reject(&is_nil/1)\n    |> case do\n      [] ->\n        typedoc\n\n      docs ->\n        \"\"\"\n        #{typedoc}\n\n        #{Enum.join(docs, \"\\n\\n\")}\n        \"\"\"\n    end\n  end\n\n  defp join_rows(rows) do\n    Enum.map_join(rows, \"\\n\", fn row -> \"|\" <> Enum.join(row, \" | \") <> \"|\" end)\n  end\nend\n```","ref":"doc_fields.html#fields"},{"type":"extras","title":"Usage - Add fields docs to the @typedoc","doc":"```elixir\ndefmodule User do\n  @moduledoc false\n\n  use TypedStructor\n\n  typed_structor do\n    @typedoc \"\"\"\n    This is a user struct.\n    \"\"\"\n    plugin Guides.Plugins.DocFields\n\n    parameter :age, doc: \"The age parameter.\"\n\n    field :name, String.t(), doc: \"The name of the user.\"\n    field :age, age, doc: \"The age of the user.\"\n  end\nend\n```\n\n```elixir\niex> t User.t\n@type t(age) :: %User{age: age | nil, name: String.t() | nil}\n\nThis is a user struct.","ref":"doc_fields.html#usage"},{"type":"extras","title":"Parameters - Add fields docs to the @typedoc","doc":"Name | Description\n:age | The age parameter.","ref":"doc_fields.html#parameters"},{"type":"extras","title":"Fields - Add fields docs to the @typedoc","doc":"Name  | Type             | Description\n:name | String.t() | nil | The name of the user.\n:age  | age | nil        | The age of the user.\n```","ref":"doc_fields.html#fields"},{"type":"extras","title":"Type Only on Ecto Schema","doc":"# Type Only on Ecto Schema\n\n`Ecto` is a great library for working with both databases and data validation.\nHowever, it has its own way of defining schemas and fields,\nwhich results a struct but without type definitions.\nThis plugin automatically disables struct creation for `Ecto` schemas.","ref":"type_only_on_ecto_schema.html"},{"type":"extras","title":"Implementation - Type Only on Ecto Schema","doc":"It uses the `c:TypedStructor.Plugin.before_definition/2` callback to determine if the module is an `Ecto` schema\nby checking the `@ecto_fields` module attribute. If it is, the `:define_struct` option is\nset to `false` to prevent struct creation.\n\nHere is the plugin(*feel free to copy and paste*):\n```elixir\ndefmodule Guides.Plugins.TypeOnlyOnEctoSchema do\n  use TypedStructor.Plugin\n\n  @impl TypedStructor.Plugin\n  defmacro before_definition(definition, _opts) do\n    quote do\n      if Module.has_attribute?(__MODULE__, :ecto_fields) do\n        Map.update!(unquote(definition), :options, fn opts ->\n          Keyword.put(opts, :define_struct, false)\n        end)\n      else\n        unquote(definition)\n      end\n    end\n  end\nend\n```","ref":"type_only_on_ecto_schema.html#implementation"},{"type":"extras","title":"Usage - Type Only on Ecto Schema","doc":"To use this plugin, you can add it to the `typed_structor` block like this:\n```elixir\ndefmodule MyApp.User do\n  use TypedStructor\n  use Ecto.Schema\n\n  typed_structor do\n    plugin Guides.Plugins.TypeOnlyOnEctoSchema\n\n    field :id, integer(), enforce: true\n    field :name, String.t()\n    field :age, integer(), enforce: true # There is always a non-nil value\n  end\n\n  schema \"source\" do\n    field :name, :string\n    field :age, :integer, default: 20\n  end\nend\n```","ref":"type_only_on_ecto_schema.html#usage"},{"type":"extras","title":"Registering the plugin globally - Type Only on Ecto Schema","doc":"```elixir\nconfig :typed_structor, plugins: [Guides.Plugins.TypeOnlyOnEctoSchema]\n```\n\nNote that the plugin is applied to **all modules** that use `TypedStructor`,\nyou can opt-out by determining the module name or other conditions.\n\nLet's change the plugin to only apply to modules from the `MyApp` namespace(*feel free to copy and paste*):\n\n```elixir\ndefmodule Guides.Plugins.TypeOnlyOnEctoSchema do\n  use TypedStructor.Plugin\n\n  @impl TypedStructor.Plugin\n  defmacro before_definition(definition, _opts) do\n    quote do\n      # Check if the module is from the MyApp namespace\n      with \"MyApp\" <- __MODULE__ |> Module.split() |> hd(),\n           true <- Module.has_attribute?(__MODULE__, :ecto_fields) do\n        Map.update!(unquote(definition), :options, fn opts ->\n          Keyword.put(opts, :define_struct, false)\n        end)\n      else\n        _otherwise -> unquote(definition)\n      end\n    end\n  end\nend\n```\n\nNow you can use `typed_structor` without registering the plugin explicitly:\n\n```diff\n defmodule MyApp.User do\n   use TypedStructor\n   use Ecto.Schema\n\n   typed_structor do\n-    plugin Guides.Plugins.TypeOnlyOnEctoSchema\n \n     field :id, integer(), enforce: true\n     field :name, String.t()\n     field :age, integer(), enforce: true\n   end\n\n   schema \"source\" do\n     field :name, :string\n     field :age, :integer, default: 20\n   end\n end\n```","ref":"type_only_on_ecto_schema.html#registering-the-plugin-globally"},{"type":"extras","title":"Add primary key and timestamps types to your Ecto schema","doc":"# Add primary key and timestamps types to your Ecto schema","ref":"primary_key_and_timestamps.html"},{"type":"extras","title":"Implementation - Add primary key and timestamps types to your Ecto schema","doc":"This plugin use `c:TypedStructor.Plugin.before_definition/2` callback to\ninject the primary key and timestamps fields to the type definition.\n\n```elixir\ndefmodule Guides.Plugins.PrimaryKeyAndTimestamps do\n  use TypedStructor.Plugin\n\n  @impl TypedStructor.Plugin\n  defmacro before_definition(definition, _opts) do\n    quote do\n      Map.update!(unquote(definition), :fields, fn fields ->\n        # Assume that the primary key is an integer\n        primary_key = [name: :id, type: quote(do: integer()), enforce: true]\n\n        # Add two default timestamps\n        timestamps = [\n          [name: :inserted_at, type: quote(do: NaiveDateTime.t()), enforce: true],\n          [name: :updated_at, type: quote(do: NaiveDateTime.t()), enforce: true]\n        ]\n\n        [primary_key | fields] ++ timestamps\n      end)\n    end\n  end\nend\n```","ref":"primary_key_and_timestamps.html#implementation"},{"type":"extras","title":"Usage - Add primary key and timestamps types to your Ecto schema","doc":"```elixir\ndefmodule User do\n  use TypedStructor\n  use Ecto.Schema\n\n  # disable struct creation or it will conflict with the Ecto schema\n  typed_structor define_struct: false do\n    # register the plugin\n    plugin Guides.Plugins.PrimaryKeyAndTimestamps\n\n    field :name, String.t()\n    field :age, integer(), enforce: true # There is always a non-nil value\n  end\n\n  schema \"source\" do\n    field :name, :string\n    field :age, :integer, default: 20\n\n    timestamps()\n  end\nend\n```\n\nIf you want to apply this plugin conditionally, refer to the [Registering plugins globally](./registering_plugins_globally.md) section.","ref":"primary_key_and_timestamps.html#usage"},{"type":"extras","title":"Derives the Jason.Encoder for struct","doc":"# Derives the `Jason.Encoder` for `struct`\n\nWe use the `c:TypedStructor.Plugin.after_definition/2` callback to\ngenerate the `Jason.Encoder` implementation for the struct.","ref":"derive_jason.html"},{"type":"extras","title":"Implementation - Derives the Jason.Encoder for struct","doc":"```elixir\ndefmodule Guides.Plugins.Jason do\n  use TypedStructor.Plugin\n\n  @impl TypedStructor.Plugin\n  defmacro after_definition(definition, _opts) do\n    quote bind_quoted: [definition: definition] do\n      # Extract the field names\n      keys = Enum.map(definition.fields, &Keyword.fetch!(&1, :name))\n\n      defimpl Jason.Encoder do\n        def encode(value, opts) do\n          Jason.Encode.map(Map.take(value, unquote(keys)), opts)\n        end\n      end\n    end\n  end\nend\n```","ref":"derive_jason.html#implementation"},{"type":"extras","title":"Usage - Derives the Jason.Encoder for struct","doc":"```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor enforce: true do\n    plugin Guides.Plugins.Jason\n\n    field :name, String.t()\n  end\nend\n```\n\nAfter compiled, you got:\n```elixir\niex> Jason.encode(%User{name: \"Phil\"})\n{:ok, \"{\\\"name\\\":\\\"Phil\\\"}\"}\n```","ref":"derive_jason.html#usage"},{"type":"extras","title":"Derives the Enumerable for struct","doc":"# Derives the `Enumerable` for `struct`\n\nWe use the `c:TypedStructor.Plugin.after_definition/2` callback to\ngenerate the `Enumerable` implementation for the struct.\nWe implement `Enumerable` callbacks exclusively using the fields that are defined.\n\nLet's start!","ref":"derive_enumerable.html"},{"type":"extras","title":"Implementation - Derives the Enumerable for struct","doc":"```elixir\ndefmodule Guides.Plugins.Enumerable do\n  use TypedStructor.Plugin\n\n  @impl TypedStructor.Plugin\n  defmacro after_definition(definition, _opts) do\n    quote bind_quoted: [definition: definition] do\n      keys = Enum.map(definition.fields, &Keyword.fetch!(&1, :name))\n\n      defimpl Enumerable do\n        def count(enumerable), do: {:ok, Enum.count(unquote(keys))}\n        def member?(enumerable, element), do: {:ok, Enum.member?(unquote(keys), element)}\n\n        def reduce(enumerable, acc, fun) do\n          # The order of fields is guaranteed to align with the sequence in which they are defined.\n          unquote(keys)\n          |> Enum.map(fn key -> {key, Map.fetch!(enumerable, key)} end)\n          |> Enumerable.List.reduce(acc, fun)\n        end\n\n        # We don't support this\n        def slice(_enumerable), do: {:error, __MODULE__}\n      end\n    end\n  end\nend\n```","ref":"derive_enumerable.html#implementation"},{"type":"extras","title":"Usage - Derives the Enumerable for struct","doc":"```elixir\ndefmodule User do\n  use TypedStructor\n\n  typed_structor enforce: true do\n    plugin Guides.Plugins.Enumerable\n\n    field :name, String.t()\n    field :age, Integer.t()\n  end\nend\n```\n\n```elixir\niex> user = %User{name: \"Phil\", age: 20}\n%User{name: \"Phil\", age: 20}\n# the order of fields is deterministic\niex> Enum.map(user, fn {key, _value} -> key end)\n[:name, :age]\n# we got a deterministic ordered Keyword list\niex> Enum.to_list(user)\n[name: \"Phil\", age: 20]\n```\n\n> #### Bonus {: .info}\n> Additionally, we gain the bonus of having a deterministic order of fields,\n> determined by the sequence in which they are defined.","ref":"derive_enumerable.html#usage"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}